#ifndef DOUBLE_INTEGER_HPP
#define DOUBLE_INTEGER_HPP

/* Double Integer template
// double_integer.hpp
//
// Author: Jan Ringos, http://Tringi.MX-3.cz, jan@ringos.cz
// Version: 1.9.2
// Description: Implements common functionality to double the precission
//              of an little-endian two-complement integer type, both signed
//              and unsigned.
//
// Changelog:
//      29.01.2011 - initial version
//      28.05.2013 - now requires C++11 (to enable singed/unsigned conversions)
//                 - fixed conversion from negative floating point
//                 - C style fmod, fmodf, fmodl replaced by std::fmod
*/

#include <type_traits>
#include <cstddef>
#include <cstdio>

// double_integer <LO, HI>
//  - unlike the intrinsic types, this one is always zero initialized
//  - arguments: LO - always an unsigned integer type
//               HI - defaults to LO, type of the upper half
//  - notes: - be very careful when initializing, di<uint> x = -1 will not work!
//           - use of types with different width for LO and HI is unsupported
//           - to implement a signed type, use a signed integer as HI type

template <typename LO, typename HI = LO>
class double_integer {
    public:
        LO lo;
        HI hi;
        
    public:
        static const std::size_t size = sizeof (LO) + sizeof (HI);

    public:
        double_integer () : lo (LO ()), hi (HI ()) {};
        double_integer (const double_integer & a) : lo (a.lo), hi (a.hi) {};
        double_integer (const LO & _lo, const HI & _hi = HI ()) : lo (_lo), hi (_hi) {};
        
    public:
        explicit double_integer (const char *, unsigned int radix = 0);
        explicit double_integer (float);
        explicit double_integer (double);
        explicit double_integer (long double);

        double_integer & operator = (const double_integer &);

        template <class = std::enable_if <!std::is_same <LO, HI>::value>>
        double_integer & operator = (const double_integer <LO, LO> &);
        
    public:
        bool operator ! () const;
        operator double_integer <LO, LO> () const;
        
        double_integer operator + () const { return *this; };
        double_integer operator - () const;
        double_integer operator ~ () const;

        double_integer operator ++ (int);
        double_integer operator -- (int);
        double_integer & operator ++ ();
        double_integer & operator -- ();

        double_integer & operator += (const double_integer &);
        double_integer & operator -= (const double_integer &);
        double_integer & operator *= (const double_integer &);
        double_integer & operator /= (const double_integer &);
        double_integer & operator %= (const double_integer &);

        double_integer & operator |= (const double_integer &);
        double_integer & operator &= (const double_integer &);
        double_integer & operator ^= (const double_integer &);

        double_integer & operator >>= (unsigned int);
        double_integer & operator <<= (unsigned int);

        double_integer div (const double_integer &, double_integer * = 0) const;
    
    private:
        template <typename T> static T mask (std::size_t);
    
    public:
        const char * print (char * buffer, std::size_t length,
                            unsigned int radix = 10u) const;
        const char * c_str (unsigned int radix = 10u) const; // only for testing
}
#ifdef __GNUC__
    __attribute__ ((__packed__))
#endif
;

#include "double_integer.tcc"
#endif
