#ifndef DOUBLE_INTEGER_TCC
#define DOUBLE_INTEGER_TCC

/* Double Integer template
// double_integer.tcc
//
// Author: Jan Ringos, http://Tringi.MX-3.cz, jan@ringos.cz
// Version: 1.9.2
//
// Changelog:
//      29.01.2011 - initial version
//      28.05.2013 - fixed conversion from negative floating point
//                 - C style fmod, fmodf, fmodl replaced by std::fmod
*/

#include <climits>
#include <cstring>
#include <cctype>
#include <cmath>

#include <limits>

template <typename LO, typename HI>
bool
double_integer <LO, HI> ::operator ! () const {
    return !(!!this->hi || !!this->lo);
};

template <typename LO, typename HI>
double_integer <LO, HI>
double_integer <LO, HI> ::operator - () const {
    double_integer <LO, HI> result = ~*this;
    ++result;
    return result;
};

template <typename LO, typename HI>
double_integer <LO, HI>
double_integer <LO, HI> ::operator ~ () const {
    return double_integer <LO, HI> (~this->lo, ~this->hi);
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator ++ () {
    ++this->lo;
    if (!this->lo) {
        ++this->hi;
    };
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator -- () {
    if (!this->lo) {
        --this->hi;
    };
    --this->lo;
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI>
double_integer <LO, HI> ::operator ++ (int) {
    double_integer <LO, HI> previous = *this;
    ++*this;
    return previous;
};

template <typename LO, typename HI>
double_integer <LO, HI>
double_integer <LO, HI> ::operator -- (int) {
    double_integer <LO, HI> previous = *this;
    --*this;
    return previous;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator += (const double_integer <LO, HI> & addend) {
    LO previous_lo = this->lo;

    this->lo += addend.lo;
    this->hi += addend.hi;
    this->hi += (this->lo < previous_lo);
    
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator -= (const double_integer <LO, HI> & subtrahend) {
    return *this += -subtrahend;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator *= (const double_integer <LO, HI> & _multiplier) {
    if (!_multiplier) {
        *this = double_integer <LO, HI> ();
        return *this;
    };

    double_integer <LO, HI> previous = *this;
    double_integer <LO, HI> multiplier = _multiplier;
    
    this->lo = LO ();
    this->hi = HI ();
    
    std::size_t i = 0u;
    std::size_t n = CHAR_BIT * double_integer <LO, HI> ::size;
    
    while (i < n && !!multiplier) {
        if (!!(multiplier.lo & LO (1u))) {
            *this += previous << i;
        };
        
        multiplier >>= 1u;
        ++i;
    };
    
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator /= (const double_integer <LO, HI> & divisor) {
    *this = this->div (divisor);
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator %= (const double_integer <LO, HI> & divisor) {
    this->div (divisor, this);
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI>
double_integer <LO, HI> ::div (const double_integer <LO, HI> & _ds,
                               double_integer <LO, HI> * remainder) const {
    
    // division by zero
    //  - propagate through to the hardware type and see what happens
    
    if (!_ds)
        return this->lo / _ds.lo;
    
    // reset sign bits
    
    double_integer <LO, HI> zero;
    double_integer <LO, HI> ds = _ds;
    double_integer <LO, HI> dd = *this;
    
    if (ds < zero) {
        ds = -ds;
    };
    if (dd < zero) {
        dd = -dd;
    };

    // remainder only
    //  - shortcut, return zero
    
    if (dd < ds) {
        if (remainder) {
            *remainder = *this;
        };
        return zero;
    };
    
    double_integer <LO, HI> r;
    double_integer <LO, HI> q;
    double_integer <LO, HI> one;
    ++one;
    
    std::size_t b = CHAR_BIT * double_integer <LO, HI> ::size;
    
    while (r < ds) {
        r <<= 1u;
        if (!!(dd & (one << --b))) {
            r.lo |= one.lo;
        };
    };
    
    while (true) {
        if (r < ds) {
            if (!b) break;
            
            --b;
            r <<= 1u;

            if (!!(dd & (one << b))) {
                r.lo |= one.lo;
            };
        } else {
            r -= ds;
            q |= one << b;
        };
    };

    // correct sign bits
    
    if ((_ds < zero) ^ (*this < zero)) {
        q = -q;
    };
    
    // results

    if (remainder) {
        if (*this < zero) {
            r = -r;
        };
        *remainder = r;
    };
    return q;
    
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator <<= (unsigned int bits) {

    // discard upper bits of the shift count, just like shift for ints does
    bits &= (CHAR_BIT * double_integer <LO, HI> ::size) - 1u;
    
    if (bits >= CHAR_BIT * sizeof (LO)) {
        bits -= CHAR_BIT * sizeof (LO);
        
        this->hi = this->lo;
        this->lo = LO ();
        this->hi <<= bits;
        
    } else {
        const unsigned int rest = CHAR_BIT * sizeof (LO) - bits;
        
        this->hi = (LO (this->hi) << bits)
                 | ((this->lo & (this->mask <LO> (bits) << rest)) >> rest);
        this->lo <<= bits;
    };
    
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator >>= (unsigned int bits) {
    
    // discard upper bits of the shift count, just like shift for ints does
    bits &= (CHAR_BIT * double_integer <LO, HI> ::size) - 1u;
    
    if (bits >= CHAR_BIT * sizeof (LO)) {
        bits -= CHAR_BIT * sizeof (LO);
        
        this->lo = this->hi;
        this->lo >>= bits;
        
        if (std::numeric_limits <HI> ::is_signed
                && double_integer (LO (), this->hi) < double_integer ()) {
            
            this->hi = ~HI ();
            this->lo |= this->mask <LO> (bits)
                    << (CHAR_BIT * sizeof (LO) - bits);
        } else {
            this->hi = HI ();
        };

    } else {
        this->lo >>= bits;
        this->lo |= (LO (this->hi) & this->mask <LO> (bits))
                 << (CHAR_BIT * sizeof (LO) - bits);
        this->hi >>= bits;
    };
    
    return *this;
};

template <typename LO, typename HI>
template <typename T>
T double_integer <LO, HI> ::mask (std::size_t n) {
    T m = T ();
    
    ++m;
    m <<= n;
    --m;
    
    return m;
};

template <typename LO, typename HI>
double_integer <LO, HI> ::double_integer (float a)
    :   lo (std::fmod (a,   static_cast <float> (LO (~LO ())) + 1.0f)),
        hi ((a < 0.0f) ? (a / (static_cast <float> (LO (~LO ())) + 1.0f) - 1u)
                       : (a / (static_cast <float> (LO (~LO ())) + 1.0f))) {};

template <typename LO, typename HI>
double_integer <LO, HI> ::double_integer (double a)
    :   lo (std::fmod (a,   static_cast <double> (LO (~LO ())) + 1.0)),
        hi ((a < 0.0) ? (a / (static_cast <double> (LO (~LO ())) + 1.0) - 1u)
                      : (a / (static_cast <double> (LO (~LO ())) + 1.0))) {};

template <typename LO, typename HI>
double_integer <LO, HI> ::double_integer (long double a)
    :   lo (std::fmod (a,   static_cast <long double> (LO (~LO ())) + 1.0l)),
        hi ((a < 0.0l) ? (a / (static_cast <long double> (LO (~LO ())) + 1.0l) - 1u)
                       : (a / (static_cast <long double> (LO (~LO ())) + 1.0l))) {};

template <typename LO, typename HI>
double_integer <LO, HI> ::double_integer (const char * s, unsigned int _radix)
    :   lo (LO ()),
        hi (HI ()) {
    
    double_integer <LO, HI> digit;
    double_integer <LO, HI> radix;
    bool minus = false;
    
    // skip blanks to first character
    while (std::isspace (*s)) ++s;

    if (*s == '-') {
        ++s;
        minus = true;
        
        // skip optional spaces after minus
        while (std::isspace (*s)) ++s;
    };
    
    // if no radix was provided, autodetect
    if (_radix) {
        radix.lo = _radix;
    } else {
        if (*s == '0') {
            ++s;
            radix.lo = 8u;
            
            if (*s == 'x' || *s == 'X') {
                ++s;
                radix.lo = 16u;
            };
            if (*s == 'b' || *s == 'B') {
                ++s;
                radix.lo = 2u;
            };
        } else {
            radix.lo = 10u;
        };
    };
    
    while (std::isxdigit (*s)) {
        if (std::isdigit (*s)) {
            digit.lo = static_cast <unsigned int> (*s - '0');
        } else
        if (std::islower (*s)) {
            digit.lo = static_cast <unsigned int> (*s - 'a' + 10u);
        } else {
            digit.lo = static_cast <unsigned int> (*s - 'A' + 10u);
        };
        
        *this *= radix;
        *this += digit;
        ++s;
    };
    
    if (minus) {
        *this = -*this;
    };
    return;
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator = (const double_integer <LO, HI> & x) {
    this->lo = x.lo;
    this->hi = x.hi;
    return *this;
};

template <typename LO, typename HI>
template <class>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator = (const double_integer <LO, LO> & x) {
    this->lo = x.lo;
    this->hi = x.hi;
    return *this;
};

template <typename LO, typename HI>
double_integer <LO, HI> ::operator double_integer <LO, LO> () const {
    return double_integer <LO, LO> (this->lo, this->hi);
};

template <typename LO, typename HI>
const char * double_integer <LO, HI> ::print (char * buffer, std::size_t length,
                                              unsigned int radix) const {
    if (!*this)
        return "0";

    if (radix < 2 || radix > 37)
        return "(inv radix)";

    if (length) {
        std::size_t i = length;
        buffer [--i] = '\0';

        double_integer <LO, HI> zero;
        double_integer <LO, HI> r;
        double_integer <LO, HI> x = *this;
        
        if (x < zero) {
            x = -x;
            if (x < zero) {
                
                buffer [0] = '-';
                double_integer <LO, LO>
                    (this->lo, this->hi) .print (buffer + 1, length - 1, radix);
                
                return buffer;
            };
        };
        
        double_integer <LO, HI> di_radix;
        di_radix.lo = radix;
        
        while (i && !!x) {
            x = x.div (di_radix, &r);
            
            unsigned char c = *reinterpret_cast <unsigned char *> (&r);
            if (c < 10u)
                buffer [--i] = c + '0';
            else
                buffer [--i] = c + 'A' - 10u;
        };

        if (*this < zero) {
            if (i) {
                buffer [--i] = '-';
            } else
                return NULL;
        };
        
        if (!x) {
            std::memmove (buffer, buffer + i, length - i);
            return buffer;
        };
    };
    
    return NULL;
};

template <typename LO, typename HI>
const char * double_integer <LO, HI> ::c_str (unsigned int radix) const {
    static char sz [CHAR_BIT * double_integer <LO, HI> ::size + 2u];
    return this->print (sz, sizeof sz, radix);
};

template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator |= (const double_integer <LO, HI> & param) {
    this->lo |= param.lo;
    this->hi |= param.hi;
    return *this;
};
template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator &= (const double_integer <LO, HI> & param) {
    this->lo &= param.lo;
    this->hi &= param.hi;
    return *this;
};
template <typename LO, typename HI>
double_integer <LO, HI> &
double_integer <LO, HI> ::operator ^= (const double_integer <LO, HI> & param) {
    this->lo ^= param.lo;
    this->hi ^= param.hi;
    return *this;
};

template <typename LO, typename HI>
bool operator < (const double_integer <LO, HI> & a,
                 const double_integer <LO, HI> & b) {
    if (a.hi == b.hi)
        return a.lo < b.lo;
    else
        return a.hi < b.hi;
};
template <typename LO, typename HI>
bool operator == (const double_integer <LO, HI> & a,
                  const double_integer <LO, HI> & b) {
    return a.lo == b.lo
        && a.hi == b.hi;
};
template <typename LO, typename HI>
bool operator != (const double_integer <LO, HI> & a,
                  const double_integer <LO, HI> & b) {
    return a.lo == b.lo
        && a.hi == b.hi;
};
template <typename LO, typename HI>
bool operator || (const double_integer <LO, HI> & a,
                  const double_integer <LO, HI> & b) {
    return (a.hi || a.lo)
        || (b.hi || b.lo);
};
template <typename LO, typename HI>
bool operator && (const double_integer <LO, HI> & a,
                  const double_integer <LO, HI> & b) {
    return (a.hi || a.lo)
        && (b.hi || b.lo);
};

// canonical operators

template <typename LO, typename HI>
double_integer <LO, HI> operator + (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) += b;
};
template <typename LO, typename HI>
double_integer <LO, HI> operator - (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) -= b;
};
template <typename LO, typename HI>
double_integer <LO, HI> operator * (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) *= b;
};
template <typename LO, typename HI>
double_integer <LO, HI> operator / (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) /= b;
};
template <typename LO, typename HI>
double_integer <LO, HI> operator % (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) %= b;
};

template <typename LO, typename HI>
double_integer <LO, HI> operator << (const double_integer <LO, HI> & a,
                                     unsigned int n) {
    return double_integer <LO, HI> (a) <<= n;
};
template <typename LO, typename HI>
double_integer <LO, HI> operator >> (const double_integer <LO, HI> & a,
                                     unsigned int n) {
    return double_integer <LO, HI> (a) >>= n;
};

template <typename LO, typename HI>
double_integer <LO, HI> operator & (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) &= b;
};
template <typename LO, typename HI>
double_integer <LO, HI> operator | (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) |= b;
};
template <typename LO, typename HI>
double_integer <LO, HI> operator ^ (const double_integer <LO, HI> & a,
                                    const double_integer <LO, HI> & b) {
    return double_integer <LO, HI> (a) ^= b;
};

template <typename LO, typename HI>
bool operator >  (const double_integer <LO, HI> & a,
                  const double_integer <LO, HI> & b) { return  (b < a); };
template <typename LO, typename HI>
bool operator <= (const double_integer <LO, HI> & a,
                  const double_integer <LO, HI> & b) { return !(b < a); };
template <typename LO, typename HI>
bool operator >= (const double_integer <LO, HI> & a,
                  const double_integer <LO, HI> & b) { return !(a < b); };

// std::numeric_limits

namespace std {
    template <>
    template <typename LO, typename HI>
    struct numeric_limits <double_integer <LO, HI> > {
        static const bool is_specialized = true;
        
        static double_integer <LO, HI> min() throw() {
            return double_integer <LO, HI> (std::numeric_limits <LO> ::min (),
                                            std::numeric_limits <HI> ::min ());
        };
        
        static double_integer <LO, HI> max() throw() {
            return double_integer <LO, HI> (std::numeric_limits <LO> ::max (),
                                            std::numeric_limits <HI> ::max ());
        };
        
        static const int digits = std::numeric_limits <HI> ::digits
                                + std::numeric_limits <LO> ::digits;
        static const int digits10 = digits * 643 / 2136;
        
        static const bool is_signed = std::numeric_limits <HI> ::is_signed;
        static const bool is_integer = true;
        static const bool is_exact = true;
        static const int radix = 2;
        static double_integer <LO, HI> epsilon() throw()
            { return double_integer <LO, HI> (); }
        static double_integer <LO, HI> round_error() throw()
            { return double_integer <LO, HI> (); }
        
        static const int min_exponent = 0;
        static const int min_exponent10 = 0;
        static const int max_exponent = 0;
        static const int max_exponent10 = 0;
        
        static const bool has_infinity = false;
        static const bool has_quiet_NaN = false;
        static const bool has_signaling_NaN = false;
        static const float_denorm_style has_denorm = denorm_absent;
        static const bool has_denorm_loss = false;
        
        static double_integer <LO, HI> infinity() throw()
            { return double_integer <LO, HI> (); }
        static double_integer <LO, HI> quiet_NaN() throw()
            { return double_integer <LO, HI> (); }
        static double_integer <LO, HI> signaling_NaN() throw()
            { return double_integer <LO, HI> (); }
        static double_integer <LO, HI> denorm_min() throw()
            { return double_integer <LO, HI> (); }
        
        static const bool is_iec559 = false;
        static const bool is_bounded = true;
        static const bool is_modulo = true;
        
        static const bool traps = std::numeric_limits <HI> ::traps;
        static const bool tinyness_before = false;
        static const float_round_style round_style = round_indeterminate;
    };
};

#endif
